<?php

namespace UnicaenDbImport\Service\Database;

use Doctrine\DBAL\Exception;
use UnicaenDbImport\Domain\Operation;
use UnicaenDbImport\Entity\Db\Service\ImportObserv\ImportObservServiceAwareTrait;
use UnicaenDbImport\Service\Exception\DatabaseServiceException;

class DatabaseService extends AbstractDatabaseService
{
    use ImportObservServiceAwareTrait;

    const SOURCE_TABLE_NAME = 'SOURCE';
    const SOURCE_TABLE_CODE_COLUMN = 'CODE';
    const DATA_SLICE_SIZE_FOR_INSERT = 1000;

    /**
     * Validation de la table destination.
     *
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function validateDestinationTable()
    {
        $connection = $this->destination->getConnection();
        $table = $this->destination->getTable();
        $sourceCodeColumn = $this->source->getSourceCodeColumn();

        // test if table exists
        $this->assertTableExists($table, $connection, "La table destination '%s' est introuvable. ");

        // id column validation
        $columnCount = 0;
        $sql = $this->codeGenerator->generateSQLForIdColumnValidationInTable($table, $columnCount);
        try {
            $result = $this->queryExecutor->fetchAll($sql, $connection);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la validation de la colonne id dans la table destination '$table'", $e);
        }
        $exception = $this->codeGenerator->convertIdColumnValidationBadResultToException($table, $result, $columnCount);
        if ($exception !== null) {
            throw DatabaseServiceException::error("La table '$table' n'est pas une table de destination valide. ", $exception);
        }

        // existence de la sequence pour la colonne id
        $idColumnSequenceName = $this->destination->getIdColumnSequence();
        if ($idColumnSequenceName !== false) {
            $sql = $this->codeGenerator->generateSQLForSequenceExistenceCheck($table, $idColumnSequenceName);
            try {
                $result = $this->queryExecutor->fetchAll($sql, $connection);
            } catch (Exception $e) {
                throw DatabaseServiceException::error("Erreur lors du test d'existence de la séquence dans la table destination '$table'", $e);
            }
            $exception = $this->codeGenerator->convertSequenceExistenceCheckResultToException($idColumnSequenceName, $result);
            if ($exception !== null) {
                throw DatabaseServiceException::error("La sequence '$idColumnSequenceName' est introuvable. ", $exception);
            }
        }

        // histo columns validation
        $columnCount = 0;
        $sql = $this->codeGenerator->generateSQLForHistoColumnsValidationInTable($table, $columnCount);
        try {
            $result = $this->queryExecutor->fetchAll($sql, $connection);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la validation des colonne d'historique dans la table destination '$table'", $e);
        }
        $exception = $this->codeGenerator->convertHistoColumnsValidationBadResultToException($table, $result, $columnCount);
        if ($exception !== null) {
            throw DatabaseServiceException::error("La table '$table' n'est pas une table de destination valide. ", $exception);
        }

        // source columns validation
        $columnCount = 0;
        $sql = $this->codeGenerator->generateSQLForSourceColumnsValidationInTable($table, $sourceCodeColumn, $columnCount);
        try {
            $result = $this->queryExecutor->fetchAll($sql, $connection);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la validation des colonne de source dans la table destination '$table'", $e);
        }
        $exception = $this->codeGenerator->convertSourceColumnsValidationBadResultToException($table, $sourceCodeColumn, $result, $columnCount);
        if ($exception !== null) {
            throw DatabaseServiceException::error("La table '$table' n'est pas une table de destination valide. ", $exception);
        }
    }

    /**
     * Validation de la table 'SOURCE' dans la destination.
     *
     * ATTENTION : on parle bien ici de la table nommée `SOURCE` qui doit exister dans la base de données destination,
     * et non pas de la table contenant les données sources de l'import/synchro.
     *
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function validateSourceTable()
    {
        $connection = $this->destination->getConnection();
        $tableName = self::SOURCE_TABLE_NAME;

        // test if table exists
        $this->assertTableExists($tableName, $connection);
    }

    /**
     * Vérifie dans la table SOURCE qu'il existe bien une source ayant le 'code' spécifié.
     *
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function validateCodeExistsInSourceTable(string $code)
    {
        $connection = $this->destination->getConnection();
        $tableName = self::SOURCE_TABLE_NAME;
        $column = self::SOURCE_TABLE_CODE_COLUMN;

        // test if value exists in table
        $sql = $this->codeGenerator->generateSQLForValueExistenceCheckInTable($column, $code, $tableName);
        try {
            $result = $this->queryExecutor->fetchAll($sql, $connection);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors du test d'existence de la valeur '$code' dans la table '$tableName'", $e);
        }
        $exists = $this->codeGenerator->convertValueExistenceCheckInTableResultToBoolean($result);
        if (!$exists) {
            throw DatabaseServiceException::error("Vous devez déclarer une source dans la table '$tableName' dont le $column est '$code'.");
        }
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function checkIntermediateTableNotExists(string $intermediateTable)
    {
        try {
            $this->assertTableExists($intermediateTable, $this->destination->getConnection());
        } catch (DatabaseServiceException $e) {
            // :-) la table n'existe pas
            return;
        }

        throw DatabaseServiceException::error(
            "Une table $intermediateTable existe déjà. Veuillez la supprimer ou alors spécifier dans la config " .
            "le nom à utiliser pour générer la table intermédiaire nécessaire à l'import dans la table " .
            $this->destination->getTable());
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function createIntermediateTable($intermediateTable)
    {
        $sql = $this->codeGenerator->generateSQLForIntermediateTableCreation($this->source, $this->destination, $intermediateTable);
        try {
            $this->queryExecutor->exec($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la création de la table intermédiaire '$intermediateTable'", $e);
        }
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function dropIntermediateTable($intermediateTable)
    {
        $sql = $this->codeGenerator->generateSQLForIntermmediateTableDrop($intermediateTable);
        try {
            $this->queryExecutor->exec($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la suppression de la table intermédiaire '$intermediateTable'", $e);
        }
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function truncateDestinationTable()
    {
        $data = $this->source->getData();
        if (empty($data)) {
            return;
        }

        $table = $this->destination->getTable();

        // détermination de l'id de la source des données à insérer, pour n'effacer que les données de cette source
        $row1 = reset($data);
        $preparedRow1 = $this->prepareSourceDataRow($row1);
        $sourceId = $preparedRow1['source_id'] ?? $preparedRow1['SOURCE_ID'];
        if (!$sourceId) {
            throw DatabaseServiceException::error(
                "Les données sources ont un 'source_id' absent ou vide, cela empêche le vidage partiel de la table destination '$table'"
            );
        }

        $sql = $this->codeGenerator->generateSQLForClearTable($table);
        $sql .= " WHERE source_id = (SELECT id FROM source WHERE code = '$sourceId')";
        try {
            $this->queryExecutor->exec($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors du vidage de la table destination '$table'", $e);
        }
    }

    /**
     * @return int
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function populateDestinationTableFromSource(): int
    {
        $destinationTable = $this->destination->getTable();
        $idColumnSequenceName = $this->destination->getIdColumnSequence();

        $connection = $this->destination->getConnection();
        try {
            $connection->beginTransaction();
            $this->truncateDestinationTable();
            $count = $this->populateTableFromSource($destinationTable, $idColumnSequenceName);
            $connection->commit();
        } catch (DatabaseServiceException $e) {
            $this->rollBack($connection);
            throw $e;
        } catch (Exception $e) {
            $this->rollBack($connection);
            throw DatabaseServiceException::error("Erreur rencontrée en bdd : " . $e->getMessage(), $e);
        }

        return $count;
    }

    /**
     * @param string $tableName
     * @param string|null|false $idColumnSequence
     * @return int
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException Aucune données dans la source ; ou anomalie ; ou erreur en bdd
     */
    public function populateTableFromSource(string $tableName, $idColumnSequence = null): int
    {
        $sourceCode = $this->source->getCode();
        $data = $this->source->getData();

        if (empty($data)) {
            return 0;
        }

        $count = 0;

        foreach ($data as $row) {
            $preparedRow = $this->prepareSourceDataRow($row);
            $columnsValues = $this->prepareDestinationData($preparedRow);

            $insertsSQL = $this->codeGenerator->generateSQLForInsertOneRowIntoTable($tableName, $columnsValues, $sourceCode, $idColumnSequence);
            try {
                $n = $this->queryExecutor->exec($insertsSQL, $this->destination->getConnection());
            } catch (Exception $e) {
                throw DatabaseServiceException::error("Erreur rencontrée lors du remplissage de la table destination '$tableName'", $e);
            }
            if ($n === 0) {
                throw DatabaseServiceException::error(sprintf(
                    "Lors du remplissage de la table destination '$tableName', un insert au moins n'a eu aucun effet, ".
                    "sans doute parce que la source dont le '%s' est '%s' n'est pas déclarée dans la table %s",
                    self::SOURCE_TABLE_CODE_COLUMN, $preparedRow['SOURCE_ID'], self::SOURCE_TABLE_NAME
                ));
            }
            $count += $n;
        }

        return $count;
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    private function prepareSourceDataRow(array $row): array
    {
        // transformations éventuelles de valeurs de colonnes/attributs
        $columnValueFilter = $this->source->getColumnValueFilter();
        if ($columnValueFilter !== null) {
            $columnValueFilter->setParams($this->source->getExtra());
            foreach ($row as $columnName => $columnValue) {
                $row[$columnName] = $columnValueFilter->filter($columnName, $columnValue);
            }
        }

        // transformations éventuelles de noms de colonnes/attributs
        $columnNameFilter = $this->source->getColumnNameFilter();
        if ($columnNameFilter !== null) {
            $filteredKeys = array_map([$columnNameFilter, 'filter'], array_keys($row));
            $row = array_combine($filteredKeys, $row);
        }

        // si la Source n'est pas fournie (par son 'code'), il faut que les données l'inclue (colonne 'SOURCE_ID')
        $sourceCode = $this->source->getCode();
        if ($sourceCode === null) {
            if (array_key_exists('source_id', $row)) {
                $row['SOURCE_ID'] = $row['source_id'];
            }
            elseif (!array_key_exists('SOURCE_ID', $row)) {
                throw DatabaseServiceException::error(
                    "Lorsque la config ne fournie pas le 'code' de la source ($this->source), " .
                    "les données doivent contenir une colonne SOURCE_ID ou source_id");
            }
        }
        else{
            $row['SOURCE_ID'] = $sourceCode;
        }

        return $row;
    }

    private function prepareDestinationData(array $row): array
    {
        $columns = $this->source->getColumns();
        $sourceCodeColumn = $this->source->getSourceCodeColumn();
        $columns = array_merge([$sourceCodeColumn], $columns);

        // préparation des données :
        // - noms de colonnes en majuscules ;
        // - mise à NULL des colonnes pour lesquelles on n'a pas de valeur.
        $columnsValues = [];
        foreach (array_map('strtoupper', $columns) as $column) {
            $columnsValues[$column] = $row[$column] ?? $row[strtolower($column)] ?? null;
        }

        // transformations éventuelles de valeurs de colonnes
        $columnValueFilter = $this->destination->getColumnValueFilter();
        if ($columnValueFilter !== null) {
            $columnValueFilter->setParams($this->source->getExtra());
            foreach ($columnsValues as $columnName => $columnValue) {
                $columnsValues[$columnName] = $columnValueFilter->filter($columnName, $columnValue);
            }
        }

        return $columnsValues;
    }

    /**
     * @param string|null $intermediateTable
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function recreateDiffView(?string $intermediateTable = null)
    {
        try {
            $sql = $this->codeGenerator->generateSQLForDiffViewCreation($this->source, $this->destination, $intermediateTable);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la génération du SQL de création de la vue différentielle", $e);
        }
        try {
            $this->queryExecutor->exec($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de la création de la vue différentielle", $e);
        }
    }

    /**
     * @return array
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function fetchDiffView(): array
    {
        $sql = $this->codeGenerator->generateSQLForDiffViewSelect($this->source, $this->destination);

        try {
            return $this->queryExecutor->fetchAll($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de l'interrogation de la vue différentielle ", $e);
        }
    }

    /**
     * @return int
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function fetchDiffViewCount(): int
    {
        $sql = $this->codeGenerator->generateSQLForDiffViewSelect($this->source, $this->destination);
        $sql = 'select count(*) as nb from (' . $sql . ') tmp';

        try {
            $result = $this->queryExecutor->fetch($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de l'interrogation de la vue différentielle ", $e);
        }

        return (int) $result['nb'];
    }

    /**
     * La table IMPORT_OBSERV permet de demander la "détection" de certains changements de valeurs
     * lors de la synchro. Les changements détectés sont inscrits dans la table IMPORT_OBSERV_RESULT.
     *
     * Cette méthode interroge la table IMPORT_OBSERV pour savoir quels changements de valeurs doivent être détectés
     * et peuple la table IMPORT_OBSERV_RESULT le cas échéant.
     *
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function populateImportObservResultTable()
    {
//        $qb = $this->importObservService->getRepository()->createQueryBuilder('io')
//            ->andWhere("upper(io.tableName) = upper(:table)")->setParameter('table', $this->destination->getTable())
//            ->andWhere("io.operation = 'UPDATE'")
//            ->andWhere("io.enabled = 1")
//            ->addOrderBy('io.operation, io.tableName, io.columnName');
//        $importObservRows = $qb->getQuery()->getResult(Query::HYDRATE_ARRAY);
        $sql = $this->codeGenerator->generateSQLForSelectingInImportObservTable($this->destination->getTable());
        try {
            $importObservRows = $this->queryExecutor->fetchAll($sql, $this->destination->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors du fetch des ImportObserv", $e);
        }

        if (!empty($importObservRows)) {
            $sql = $this->codeGenerator->generateSQLForInsertionIntoImportObservResult($this->destination, $importObservRows);
            try {
                $this->queryExecutor->exec($sql, $this->destination->getConnection());
            } catch (Exception $e) {
                throw DatabaseServiceException::error("Erreur lors de la création des ImportObservResult", $e);
            }
        }
    }

    /**
     * @return array [operation => int|Exception]
     *
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function updateDestination(): array
    {
        $connection = $this->destination->getConnection();
        $results = [];

        foreach (Operation::OPERATIONS as $operation) {
            $sql = $this->codeGenerator->generateSQLForDestinationUpdate($operation, $this->source, $this->destination);
            try {
                $connection->beginTransaction();
                $result = $this->queryExecutor->exec($sql, $connection);
                $connection->commit();
            } catch (Exception $e) {
                $this->rollBack($connection);
                $result = DatabaseServiceException::error("Erreur lors de l'opération '$operation'", $e);
            }

            $results[$operation] = $result;
        }

        return $results;
    }

    /**
     * Requête la Source pour obtenir les données sources.
     *
     * @return array
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function fetchSource(): array
    {
        $selectSql = $this->codeGenerator->generateSQLForSelectFromSource($this->source);

        try {
            return $this->queryExecutor->fetchAll($selectSql, $this->source->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors de l'interrogation de la table/select", $e);
        }
    }

    /**
     * Requête la Source pour obtenir le 1er enregistrement des données sources.
     *
     * @return array|null
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function fetchSourceFirstRow(): ?array
    {
        $selectSql = $this->codeGenerator->generateSQLForSelectFromSource($this->source);

        try {
            return $this->queryExecutor->fetch($selectSql, $this->source->getConnection());
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors du fetch", $e);
        }
    }

    /**
     * @throws \UnicaenDbImport\Service\Exception\DatabaseServiceException
     */
    public function fetchSourceTableColumnNames(): array
    {
        $sourceTableName = $this->source->getTable();
        $selectSql = $this->codeGenerator->generateSQLForSelectSourceTableColumnsInformation($sourceTableName);

        try {
            $result = $this->queryExecutor->fetchAll($selectSql, $this->source->getConnection());
            if (!$result) {
                throw DatabaseServiceException::error("Impossible de déterminer les colonnes de la table/vue source '$sourceTableName'");
            }
            return array_map(fn ($row) => $row['column_name'], $result);
        } catch (Exception $e) {
            throw DatabaseServiceException::error("Erreur lors du fetch", $e);
        }
    }
}